package chizap

import (
	"net/http"
	"time"

	"github.com/go-chi/chi/v5"
	"github.com/go-chi/chi/v5/middleware"
	"go.uber.org/zap"
	"go.uber.org/zap/zapcore"
)

// Logger returns a go-chi middleware that logs requests using go.uber.org/zap.
func Logger(logger *zap.Logger, opts ...Option) func(next http.Handler) http.Handler {
	cfg := newConfig(opts...)
	return func(next http.Handler) http.Handler {
		return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
			for _, f := range cfg.filters {
				if !f(r) {
					next.ServeHTTP(w, r)
					return
				}
			}

			path := r.URL.Path
			query := r.URL.RawQuery
			ww := middleware.NewWrapResponseWriter(w, r.ProtoMajor)
			start := time.Now()

			next.ServeHTTP(ww, r)

			end := time.Now()
			statusCode := ww.Status()

			fields := []zap.Field{
				zap.Int("statusCode", statusCode),
				zap.Duration("duration", end.Sub(start)),
				zap.String("ip", cfg.ipFn(r)),
				zap.String("method", r.Method),
				zap.String("query", query),
				zap.String("path", path),
				zap.String("proto", r.Proto),
				zap.String("referer", r.Referer()),
				zap.Int("contentSize", ww.BytesWritten()),
				zap.String("userAgent", r.UserAgent()),
				zap.String("routePattern", chi.RouteContext(r.Context()).RoutePattern()),
			}

			for _, f := range cfg.additionalFieldExtractors {
				fields = append(fields, f(r)...)
			}

			logger.Log(statusCodeToLogLvl(statusCode), path, fields...)
		})
	}
}

func statusCodeToLogLvl(statusCode int) zapcore.Level {
	if statusCode >= http.StatusInternalServerError {
		return zap.ErrorLevel
	}

	if statusCode >= http.StatusBadRequest {
		return zap.WarnLevel
	}

	return zap.InfoLevel
}
